import json
import math
import os
from pathlib import Path
from typing import List, Union

import cmocean as cm
import matplotlib
import numpy as np
import pandas
from matplotlib import pyplot as plt
from shapely.geometry import LineString

from polarityjam.compute.shape import get_divisor_lines
from polarityjam.model.collection import PropertiesCollection
from polarityjam.model.image import BioMedicalChannel
from polarityjam.model.masks import (BioMedicalInstanceSegmentationMask,
                                     BioMedicalMask)
from polarityjam.model.parameter import ImageParameter, PlotParameter
from polarityjam.polarityjam_logging import get_logger
from polarityjam.vizualization.plot import (add_colorbar, add_scalebar,
                                            add_title, add_vector,
                                            save_current_fig)


class Plotter:
    """Plotter class"""

    def __init__(self, params: PlotParameter):
        self.params = params
        self.set_figure_dpi()

    def set_figure_dpi(self):
        """Set figure dpi"""
        matplotlib.rcParams["figure.dpi"] = self.params.dpi

    def _get_figure(self, n_subfigures: int):
        w, h = self.params.graphics_width, self.params.graphics_height
        fig, ax = plt.subplots(1, n_subfigures, figsize=(w * n_subfigures, h))
        plt.tight_layout()

        return fig, ax

    def _get_inlines(
            self,
            im_marker: BioMedicalChannel,
            cell_mask: BioMedicalInstanceSegmentationMask,
            nuclei_mask: BioMedicalInstanceSegmentationMask,
            single_cell_dataset: pandas.DataFrame,
    ) -> List[np.ndarray]:
        inlines_cell = BioMedicalInstanceSegmentationMask.empty(im_marker.data.shape)
        inlines_mem = np.zeros((im_marker.data.shape[0], im_marker.data.shape[1]))
        inlines_nuc = BioMedicalInstanceSegmentationMask.empty(im_marker.data.shape)

        for cell_label in single_cell_dataset["label"]:
            feature_row = single_cell_dataset.loc[
                single_cell_dataset["label"] == cell_label
                ]
            intensity_cell = feature_row["marker_mean_expression"].values[0]
            intensity_mem = feature_row["marker_mean_expression_mem"].values[0]

            intensity_nuc = None
            if nuclei_mask is not None:
                intensity_nuc = feature_row["marker_mean_expression_nuc"].values[0]

            single_cell_mask = cell_mask.get_single_instance_maks(cell_label)
            invert_outline_cell = (
                single_cell_mask.get_outline_from_mask(self.params.outline_width)
                .invert()
                .to_instance_mask()
            )
            single_cell_inlay_mask = single_cell_mask.overlay_instance_segmentation(
                invert_outline_cell
            )
            single_cell_inlay_intensity_mask = single_cell_inlay_mask.scalar_mult(
                intensity_cell
            )
            inlines_cell = inlines_cell.element_add(single_cell_inlay_intensity_mask)

            # membrane outline cannot be summed up on an empty mask, because outlines overlap.
            outline_mem = single_cell_mask.get_outline_from_mask(
                self.params.membrane_thickness
            )
            inlines_mem = np.where(
                np.logical_and(outline_mem.data, inlines_mem.data < intensity_mem),
                intensity_mem,
                inlines_mem.data,
            )

            # nuclei marker intensity
            if nuclei_mask is not None:
                single_nucleus_mask = nuclei_mask.get_single_instance_maks(cell_label)
                invert_outline_nuc = (
                    single_nucleus_mask.get_outline_from_mask(self.params.outline_width)
                    .invert()
                    .to_instance_mask()
                )
                single_nuc_inlay_mask = (
                    single_nucleus_mask.overlay_instance_segmentation(
                        invert_outline_nuc
                    )
                )
                single_nuc_inlay_intensity_mask = single_nuc_inlay_mask.scalar_mult(
                    intensity_nuc
                )
                inlines_nuc = inlines_nuc.element_add(single_nuc_inlay_intensity_mask)

        return [inlines_cell.data, inlines_mem, inlines_nuc.data]

    def _masked_cell_outlines(
            self,
            channel: BioMedicalChannel,
            instance_seg_mask: BioMedicalInstanceSegmentationMask,
    ) -> np.ndarray:
        # cell outlines
        outlines_cells = BioMedicalMask.empty(channel.data.shape)
        for cell_label in instance_seg_mask.get_labels():
            single_cell_mask = instance_seg_mask.get_single_instance_maks(cell_label)
            outline_cell = single_cell_mask.get_outline_from_mask(
                self.params.outline_width
            )
            outlines_cells = outlines_cells.operation(outline_cell, np.logical_or)

        # convert cell outlines to image
        outlines_cells_rgba = outlines_cells.to_instance_mask().mask_background()
        outlines_cells_rgba = np.dstack([outlines_cells_rgba.data] * 3)

        return outlines_cells_rgba

    def plot_channels(
            self,
            seg_img: np.ndarray,
            seg_img_params: ImageParameter,
            output_path: Union[str, Path],
            filename: Union[str, Path],
            close=False,
    ):
        """Plots the separate channels from the input image given, based on its parameters.

        Args:
            seg_img:
                numpy array of the image to plot
            seg_img_params:
                parameters of the image to plot
            output_path:
                path to the output directory where plots are saved
            filename:
                name of the file to save
            close:
                whether to close the figure after saving

        """
        get_logger().info("Plotting: input channels")

        filename, _ = os.path.splitext(os.path.basename(filename))

        if (
                seg_img_params.channel_junction is not None
                and seg_img_params.channel_nucleus is not None
        ):
            fig, ax = self._get_figure(2)

            # junction channel
            c_junction = seg_img_params.channel_junction
            ax[0].imshow(seg_img[c_junction, :, :])
            add_title(
                ax[0],
                "junction channel",
                seg_img[c_junction, :, :],
                self.params.show_graphics_axis,
            )

            # nucleus channel
            c_nucleus = seg_img_params.channel_nucleus
            ax[1].imshow(seg_img[c_nucleus, :, :])
            add_title(
                ax[1],
                "nuclei channel",
                seg_img[c_nucleus, :, :],
                self.params.show_graphics_axis,
            )

            axes = [ax[0], ax[1]]
        else:
            fig, ax = self._get_figure(1)

            # first channel
            ax.imshow(seg_img[:, :])
            add_title(
                ax, "first channel", seg_img[:, :], self.params.show_graphics_axis
            )
            axes = [ax]

        self._finish_plot(
            fig,
            output_path,
            filename,
            "_channels",
            axes,
            seg_img_params.pixel_to_micron_ratio,
            close,
        )

        return fig, axes

    def plot_mask(
            self,
            mask: np.ndarray,
            seg_img: np.ndarray,
            seg_img_params: ImageParameter,
            output_path: Union[str, Path],
            filename: Union[str, Path],
            close: bool = False,
    ):
        """Plots the segmentation mask, together with the separate channels from the input image.


        Args:
            mask:
                numpy array of the mask to plot
            seg_img:
                numpy array of the image to plot
            seg_img_params:
                parameters of the image to plot
            output_path:
                path to the output directory where plots are saved
            filename:
                name of the file to save
            close:
                whether to close the figure after saving

        """
        get_logger().info("Plotting: segmentation masks")

        filename, _ = os.path.splitext(os.path.basename(filename))

        # color each cell differently
        cell_idx = np.unique(mask)
        cell_idx = np.delete(cell_idx, 0)
        mask_ = np.copy(mask)

        new_col = np.copy(cell_idx)
        np.random.seed(42)  # set seed for reproducibility
        np.random.shuffle(new_col)
        for i in range(len(cell_idx)):
            mask_[mask == cell_idx[i]] = new_col[i]

        # ignore background
        mask_ = np.where(mask > 0, mask_, np.nan)

        if (
                seg_img_params.channel_junction is not None
                and seg_img_params.channel_nucleus is not None
        ):
            fig, ax = self._get_figure(3)

            ax[0].imshow(seg_img[0, :, :])
            add_title(
                ax[0],
                "junction channel",
                seg_img[0, :, :],
                self.params.show_graphics_axis,
            )

            ax[1].imshow(seg_img[1, :, :])
            add_title(
                ax[1],
                "nuclei channel",
                seg_img[1, :, :],
                self.params.show_graphics_axis,
            )

            ax[2].imshow(seg_img[0, :, :])
            ax[2].imshow(mask_, cmap=plt.cm.gist_rainbow, alpha=0.5)
            add_title(
                ax[2], "segmentation", seg_img[0, :, :], self.params.show_graphics_axis
            )

            axes = [ax[0], ax[1], ax[2]]
        else:
            fig, ax = self._get_figure(2)

            s_img = seg_img[:, :]

            # ax 1
            ax[0].imshow(s_img)
            add_title(ax[0], "junction channel", s_img, self.params.show_graphics_axis)

            # ax 2
            ax[1].imshow(s_img)
            ax[1].imshow(mask_, cmap=plt.cm.gist_rainbow, alpha=0.5)
            add_title(ax[1], "segmentation", s_img, self.params.show_graphics_axis)

            axes = [ax[0], ax[1]]

        self._finish_plot(
            fig,
            output_path,
            filename,
            "_segmentation",
            axes,
            seg_img_params.pixel_to_micron_ratio,
            close,
        )

        return fig, axes

    def plot_organelle_polarity(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the organelle polarity of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_junction = img.junction.data
        con_inst_seg_mask = img.segmentation.segmentation_mask_connected
        inst_nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg")
        inst_organelle_mask = img.organelle.get_mask_by_name("organelle_mask_seg")

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: organelle polarity")

        fig, ax = self._get_figure(1)

        # resources image
        ax.imshow(im_junction, cmap=plt.cm.gray, alpha=1.0)

        # determine polarity_angle
        polarity_angle_vec = collection.get_properties_by_img_name(img_name)[
            "organelle_orientation_deg"
        ].values
        polarity_angle = con_inst_seg_mask.relabel(polarity_angle_vec)

        # plot polarity angle
        cax = ax.imshow(
            polarity_angle.mask_background().data,
            cmap=cm.cm.phase,
            vmin=0,
            vmax=360,
            alpha=0.5,
        )
        color_bar = fig.colorbar(cax, ax=ax, shrink=0.3)  # , extend='both')
        color_bar.set_label("polarity angle")
        color_bar.ax.set_yticks([0, 90, 180, 270, 360])

        # plot differently colored organelle (red) and nuclei (blue)
        zero = np.zeros((im_junction.shape[0], im_junction.shape[1]))
        rgba_organelle = np.dstack(
            (
                inst_organelle_mask.to_semantic_mask().data,
                zero,
                zero,
                inst_organelle_mask.to_semantic_mask().data,
            )
        )
        rgba_nuclei = np.dstack(
            (
                zero,
                zero,
                inst_nuclei_mask.to_semantic_mask().data,
                inst_nuclei_mask.to_semantic_mask().data,
            )
        )
        ax.imshow(rgba_nuclei)
        ax.imshow(rgba_organelle)

        # plot polarity vector
        for _, row in collection.get_properties_by_img_name(img_name).iterrows():
            add_vector(
                ax,
                row["nuc_X"],
                row["nuc_Y"],
                row["organelle_X"],
                row["organelle_Y"],
                self.params.marker_size,
                self.params.font_color,
            )
            if self.params.show_polarity_angles:
                ax.text(
                    row["cell_X"],
                    row["cell_Y"],
                    str(int(np.round(row["organelle_orientation_deg"], 0))),
                    color=self.params.font_color,
                    fontsize=self.params.fontsize_text_annotations,
                )

        # set title and ax limits
        add_title(
            ax,
            "organelle orientation",
            im_junction.data,
            self.params.show_graphics_axis,
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_nuclei_organelle_vector",
            [ax],
            pixel_to_micron_ratio,
            close,
            polarity_angle,
        )

        return fig, [ax]

    def plot_nuc_displacement_orientation(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the nucleus displacement orientation of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: marker nucleus polarity")

        # figure and axes
        fig, ax = self._get_figure(1)

        # resources image
        ax.imshow(img.junction.data, cmap=plt.cm.gray, alpha=1.0)

        # determine nucleus polarity_angle
        nuc_polarity_angle_vec = collection.get_properties_by_img_name(img_name)[
            "nuc_displacement_orientation_deg"
        ].values
        nuc_polarity_angle = img.segmentation.segmentation_mask.relabel(
            nuc_polarity_angle_vec
        )

        # plot polarity angle
        cax = ax.imshow(
            nuc_polarity_angle.mask_background().data,
            cmap=cm.cm.phase,
            vmin=0,
            vmax=360,
            alpha=0.5,
        )
        color_bar = fig.colorbar(cax, ax=ax, shrink=0.3)  # , extend='both')
        color_bar.set_label("polarity angle")
        color_bar.ax.set_yticks([0, 90, 180, 270, 360])

        nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg").to_semantic_mask()

        # plot nuclei (blue)
        zero = np.zeros((img.junction.data.shape[0], img.junction.data.shape[1]))
        rgba_nuclei = np.dstack((zero, zero, nuclei_mask.data, nuclei_mask.data))
        ax.imshow(rgba_nuclei)

        # plot polarity vector
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            add_vector(
                ax,
                row["cell_X"],
                row["cell_Y"],
                row["nuc_X"],
                row["nuc_Y"],
                self.params.marker_size,
                self.params.font_color,
            )
            if self.params.show_polarity_angles:
                ax.text(
                    row["nuc_X"],
                    row["nuc_Y"],
                    str(int(np.round(row["nuc_displacement_orientation_deg"], 0))),
                    color=self.params.font_color,
                    fontsize=self.params.fontsize_text_annotations,
                )

        # set title and ax limits
        add_title(
            ax,
            "nucleus displacement orientation",
            img.junction.data,
            self.params.show_graphics_axis,
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_nucleus_displacement_orientation",
            [ax],
            pixel_to_micron_ratio,
            close,
            nuc_polarity_angle,
        )

        return fig, [ax]

    def plot_marker_expression(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the marker expression of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_marker = img.marker.data
        cell_mask = img.segmentation.segmentation_mask_connected
        single_cell_dataset = collection.dataset.loc[
            collection.dataset["filename"] == img_name
            ]

        nuclei_mask = None
        if img.img_params.channel_nucleus >= 0:
            nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg")

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: marker expression")
        # figure and axes
        number_sub_figs = 2  # mean intensity cell, mean intensity membrane
        if nuclei_mask is not None:
            number_sub_figs = 3  # (optional) mean intensity nucleus

        fig, ax = self._get_figure(number_sub_figs)

        # plot marker intensity for all subplots
        for i in range(number_sub_figs):
            ax[i].imshow(im_marker, cmap=plt.cm.gray, alpha=1.0)

        outlines_cell, outlines_mem, outlines_nuc = self._get_inlines(
            im_marker, cell_mask, nuclei_mask, single_cell_dataset
        )

        # cell and membrane outline
        outlines_cell_ = np.where(outlines_cell > 0, outlines_cell, np.nan)
        cax_1 = ax[0].imshow(outlines_cell_, plt.cm.bwr, alpha=0.5)

        outlines_mem_ = np.where(outlines_mem > 0, outlines_mem, np.nan)
        cax_2 = ax[1].imshow(outlines_mem_, plt.cm.bwr, alpha=0.5)

        # nuclei marker intensity
        cax_3 = None
        outlines_nuc_ = None
        if nuclei_mask is not None:
            outlines_nuc_ = np.where(outlines_nuc > 0, outlines_nuc, np.nan)
            cax_3 = ax[2].imshow(
                outlines_nuc_, plt.cm.bwr, alpha=0.75
            )  # always last axis

        # colorbar for cell
        yticks_cell = [
            np.nanmin(outlines_cell_),
            np.nanmax(
                outlines_cell_,
            ),
        ]
        add_colorbar(fig, cax_1, ax[0], yticks_cell, "intensity cell")

        # colorbar for membrane
        yticks_mem = [
            np.nanmin(outlines_mem_),
            np.nanmax(
                outlines_mem_,
            ),
        ]
        add_colorbar(fig, cax_2, ax[1], yticks_mem, "intensity membrane")

        # colorbar for nucleus
        if nuclei_mask is not None:
            yticks_nuc = [np.nanmin(outlines_nuc_), np.nanmax(outlines_nuc_)]
            add_colorbar(fig, cax_3, ax[2], yticks_nuc, "intensity nucleus")

        # plot mean expression value of cell and membrane as text
        for index, row in single_cell_dataset.iterrows():
            ax[0].text(
                row["cell_X"],
                row["cell_Y"],
                str(np.round(row["marker_mean_expression"], 1)),
                color=self.params.font_color,
                fontsize=self.params.fontsize_text_annotations,
            )
            ax[1].text(
                row["cell_X"],
                row["cell_Y"],
                str(np.round(row["marker_mean_expression_mem"], 1)),
                color=self.params.font_color,
                fontsize=self.params.fontsize_text_annotations,
            )
            if nuclei_mask is not None:
                ax[2].text(
                    row["nuc_X"],
                    row["nuc_Y"],
                    str(np.round(row["marker_mean_expression_nuc"], 1)),
                    color=self.params.font_color,
                    fontsize=self.params.fontsize_text_annotations,
                )

        # set title
        axes = [ax[0], ax[1]]
        add_title(
            ax[0], "mean intensity cell", im_marker.data, self.params.show_graphics_axis
        )
        add_title(
            ax[1],
            "mean intensity membrane",
            im_marker.data,
            self.params.show_graphics_axis,
        )
        if nuclei_mask is not None:
            add_title(
                ax[2],
                "mean intensity nucleus",
                im_marker.data,
                self.params.show_graphics_axis,
            )
            axes = [ax[0], ax[1], ax[2]]

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_marker_expression",
            axes,
            pixel_to_micron_ratio,
            close,
        )

        return fig, axes

    def plot_marker_polarity(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the marker polarity of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_marker = img.marker
        cell_mask = img.segmentation.segmentation_mask_connected

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: marker polarity")

        # figure and axes
        fig, ax = self._get_figure(1)

        # plot marker intensity
        ax.imshow(im_marker.data, cmap=plt.cm.gray, alpha=1.0)

        # show cell outlines
        ax.imshow(self._masked_cell_outlines(im_marker, cell_mask), alpha=0.5)

        # add all polarity vectors
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            add_vector(
                ax,
                row["cell_X"],
                row["cell_Y"],
                row["marker_centroid_X"],
                row["marker_centroid_Y"],
                self.params.marker_size,
                self.params.font_color,
            )

        add_title(ax, "marker polarity", im_marker.data, self.params.show_graphics_axis)

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_marker_polarity",
            [ax],
            pixel_to_micron_ratio,
            close,
        )

        return fig, [ax]

    def plot_marker_nucleus_orientation(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the marker polarity of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_junction = img.junction.data
        segmentation_mask = img.segmentation.segmentation_mask_connected
        inst_nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg")

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: marker nucleus polarity")

        # figure and axes
        fig, ax = self._get_figure(1)

        # resources image
        ax.imshow(im_junction, cmap=plt.cm.gray, alpha=1.0)

        # determine nucleus polarity_angle
        marker_nucleus_orientation_deg_vec = collection.get_properties_by_img_name(
            img_name
        )["marker_nucleus_orientation_deg"].values
        nuc_polarity_angle = segmentation_mask.relabel(
            marker_nucleus_orientation_deg_vec
        )

        # plot polarity angle
        cax = ax.imshow(
            nuc_polarity_angle.mask_background().data,
            cmap=cm.cm.phase,
            vmin=0,
            vmax=360,
            alpha=0.5,
        )
        color_bar = fig.colorbar(cax, ax=ax, shrink=0.3)  # , extend='both')
        color_bar.set_label("polarity angle")
        color_bar.ax.set_yticks([0, 90, 180, 270, 360])

        # plot nuclei (blue)
        zero = np.zeros((im_junction.shape[0], im_junction.shape[1]))
        rgba_nuclei = np.dstack(
            (
                zero,
                zero,
                inst_nuclei_mask.to_semantic_mask().data,
                inst_nuclei_mask.to_semantic_mask().data,
            )
        )
        ax.imshow(rgba_nuclei)

        # plot polarity vector
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            add_vector(
                ax,
                row["nuc_X"],
                row["nuc_Y"],
                row["marker_centroid_X"],
                row["marker_centroid_Y"],
                self.params.marker_size,
                self.params.font_color,
            )
            if self.params.show_polarity_angles:
                ax.text(
                    row["nuc_X"],
                    row["nuc_Y"],
                    str(int(np.round(row["marker_nucleus_orientation_deg"], 0))),
                    color=self.params.font_color,
                    fontsize=self.params.fontsize_text_annotations,
                )

        # set title and ax limits
        add_title(
            ax,
            "marker nucleus orientation",
            im_junction.data,
            self.params.show_graphics_axis,
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_marker_nucleus_orientation",
            [ax],
            pixel_to_micron_ratio,
            close,
            nuc_polarity_angle,
        )

        return fig, [ax]

    def plot_junction_polarity(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the junction polarity of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_junction = img.junction
        cell_mask = img.segmentation.segmentation_mask_connected

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: junction polarity")

        # figure and axes
        fig, ax = self._get_figure(1)

        # plot marker intensity
        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        # show cell outlines
        ax.imshow(self._masked_cell_outlines(im_junction, cell_mask), alpha=0.5)

        # add all polarity vectors
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            add_vector(
                ax,
                row["cell_X"],
                row["cell_Y"],
                row["junction_centroid_X"],
                row["junction_centroid_Y"],
                self.params.marker_size,
                self.params.font_color,
            )

        add_title(
            ax, "junction polarity", im_junction.data, self.params.show_graphics_axis
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_junction_polarity",
            [ax],
            pixel_to_micron_ratio,
            close,
        )

        return fig, [ax]

    def plot_corners(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the corners of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        fig, ax = self._get_figure(1)

        img = collection.get_image_by_img_name(img_name)
        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        # plot marker intensity
        im_junction = img.junction
        cell_mask = img.segmentation.segmentation_mask_connected

        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        for index, row in collection.dataset.loc[
            collection.dataset["filename"] == img_name
        ].iterrows():
            plt.scatter(
                np.array(json.loads(row["cell_corner_points"]))[:, 0],
                np.array(json.loads(row["cell_corner_points"]))[:, 1],
                [4] * len(np.array(json.loads(row["cell_corner_points"]))[:, 1]),
            )

        ax.imshow(self._masked_cell_outlines(im_junction, cell_mask), alpha=0.5)

        add_title(ax, "cell corners", im_junction.data, self.params.show_graphics_axis)

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_cell_corners",
            [ax],
            pixel_to_micron_ratio,
            close,
        )

        return fig, [ax]

    def plot_eccentricity(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the eccentricity of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)

        im_junction = img.junction
        segmentation_mask = img.segmentation.segmentation_mask_connected
        inst_nuclei_mask = None
        if img.img_params.channel_nucleus >= 0:
            inst_nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg")

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: eccentricity")

        # figure and axes
        number_sub_figs = 1
        if inst_nuclei_mask is not None:
            number_sub_figs = 2

        fig, ax = self._get_figure(number_sub_figs)

        # get cell_eccentricity
        cell_eccentricity_vec = collection.get_properties_by_img_name(img_name)[
            "cell_eccentricity"
        ].values
        cell_eccentricity = segmentation_mask.relabel(cell_eccentricity_vec)

        # add cell (and nuclei) eccentricity to the figure
        if inst_nuclei_mask is not None:
            Plotter._add_cell_eccentricity(fig, ax[0], im_junction, cell_eccentricity)
            # get nuclei eccentricity
            nuclei_eccentricity_vec = collection.get_properties_by_img_name(img_name)[
                "nuc_eccentricity"
            ].values
            nuclei_eccentricity = inst_nuclei_mask.relabel(nuclei_eccentricity_vec)

            get_logger().info(
                "Maximal nuclei eccentricity: %s"
                % str(np.max(nuclei_eccentricity.data))
            )
            get_logger().info(
                "Minimal nuclei eccentricity: %s"
                % str(np.min(nuclei_eccentricity.data))
            )

            Plotter._add_nuclei_eccentricity(
                fig, ax[1], im_junction, nuclei_eccentricity
            )
        else:
            Plotter._add_cell_eccentricity(fig, ax, im_junction, cell_eccentricity)

        # plot major and minor axis
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            if inst_nuclei_mask is not None:
                # plot orientation degree
                Plotter._add_single_cell_eccentricity_axis(
                    ax[0],
                    row["cell_X"],
                    row["cell_Y"],
                    row["cell_shape_orientation_rad"],
                    row["cell_major_axis_length"],
                    row["cell_minor_axis_length"],
                    row["cell_eccentricity"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )

                # plot orientation degree nucleus
                Plotter._add_single_cell_eccentricity_axis(
                    ax[1],
                    row["nuc_X"],
                    row["nuc_Y"],
                    row["nuc_shape_orientation_rad"],
                    row["nuc_major_axis_length"],
                    row["nuc_minor_axis_length"],
                    row["nuc_eccentricity"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )
            else:
                Plotter._add_single_cell_eccentricity_axis(
                    ax,
                    row["cell_X"],
                    row["cell_Y"],
                    row["cell_shape_orientation_rad"],
                    row["cell_major_axis_length"],
                    row["cell_minor_axis_length"],
                    row["cell_eccentricity"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )

        # set title and ax limits
        if inst_nuclei_mask is not None:
            add_title(
                ax[0],
                "cell elongation",
                im_junction.data,
                self.params.show_graphics_axis,
            )
            add_title(
                ax[1],
                "nuclei elongation",
                im_junction.data,
                self.params.show_graphics_axis,
            )
            axes = [ax[0], ax[1]]
        else:
            add_title(
                ax, "cell elongation", im_junction.data, self.params.show_graphics_axis
            )
            axes = [ax]

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_eccentricity",
            axes,
            pixel_to_micron_ratio,
            close,
        )

        return fig, axes

    def plot_marker_cue_intensity_ratio(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the marker cue intensity ratios of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        get_logger().info("Plotting: marker cue intensity ratios")

        img = collection.get_image_by_img_name(img_name)
        params = collection.get_runtime_params_by_img_name(img_name)

        im_junction = img.junction
        cell_mask = img.segmentation.segmentation_mask_connected

        mcdir = collection.get_properties_by_img_name(img_name)[
            "marker_cue_directional_intensity_ratio"
        ].values
        mcdir_mask = cell_mask.relabel(mcdir)

        mcuir = collection.get_properties_by_img_name(img_name)[
            "marker_cue_undirectional_intensity_ratio"
        ].values
        mcuir_mask = cell_mask.relabel(mcuir)

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        ax, fig = self._plot_cue_intensity_ratio(
            cell_mask, collection, im_junction, img_name, mcdir_mask, mcuir_mask, params
        )

        add_title(
            ax[0],
            "marker cue directional intensity ratio",
            im_junction.data,
            self.params.show_graphics_axis,
        )
        add_title(
            ax[1],
            "marker cue undirectional intensity ratio",
            im_junction.data,
            self.params.show_graphics_axis,
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_marker_ratio_method",
            ax,
            pixel_to_micron_ratio,
            close,
        )

    def plot_junction_cue_intensity_ratio(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the junction cue intensity ratios of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        get_logger().info("Plotting: junction cue intensity ratios")

        img = collection.get_image_by_img_name(img_name)
        params = collection.get_runtime_params_by_img_name(img_name)

        im_junction = img.junction
        cell_mask = img.segmentation.segmentation_mask_connected

        jcdir = collection.get_properties_by_img_name(img_name)[
            "junction_cue_directional_intensity_ratio"
        ].values
        jcdir_mask = cell_mask.relabel(jcdir)

        jcuir = collection.get_properties_by_img_name(img_name)[
            "junction_cue_undirectional_intensity_ratio"
        ].values
        jcuir_mask = cell_mask.relabel(jcuir)

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        ax, fig = self._plot_cue_intensity_ratio(
            cell_mask, collection, im_junction, img_name, jcdir_mask, jcuir_mask, params
        )

        add_title(
            ax[0],
            "junction cue directional intensity ratio",
            im_junction.data,
            self.params.show_graphics_axis,
        )
        add_title(
            ax[1],
            "junction cue undirectional intensity ratio",
            im_junction.data,
            self.params.show_graphics_axis,
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_junction_ratio_method",
            ax,
            pixel_to_micron_ratio,
            close,
        )

        return fig, ax

    def _plot_cue_intensity_ratio(
            self,
            cell_mask,
            collection,
            im_junction,
            img_name,
            directional_mask,
            undirectional_mask,
            params,
    ):
        # figure and axes
        fig, ax = self._get_figure(2)
        # show junction and cell mask overlay
        ax[0].imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)
        cax1 = ax[0].imshow(directional_mask.mask_background().data, cmap=cm.cm.balance, alpha=0.5)
        ax[1].imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)
        cax2 = ax[1].imshow(undirectional_mask.mask_background().data, cmap=cm.cm.balance, alpha=0.5)
        # show cell outlines
        ax[0].imshow(self._masked_cell_outlines(im_junction, cell_mask), alpha=0.5)
        ax[1].imshow(self._masked_cell_outlines(im_junction, cell_mask), alpha=0.5)
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            x0 = row["cell_X"]
            y0 = row["cell_Y"]

            # plot center of cell
            ax[0].plot(x0, y0, ".b", markersize=5)
            ax[1].plot(x0, y0, ".b", markersize=5)

            a = [x0, y0]
            b = [x0 + row["cell_major_axis_length"], y0]  # lies horizontally
            ground_line = LineString([a, b])

            d_lines, _ = get_divisor_lines(a, params.cue_direction, ground_line, 4)
            for d_line in d_lines:
                x1, y1 = (i[0] for i in d_line.boundary.centroid.coords.xy)
                ax[1].plot((x0, x1), (y0, y1), "--r", linewidth=0.5)

            d_lines, _ = get_divisor_lines(a, params.cue_direction, ground_line, 2)
            for d_line in d_lines:
                x1, y1 = (i[0] for i in d_line.boundary.centroid.coords.xy)
                ax[0].plot((x0, x1), (y0, y1), "--r", linewidth=0.5)

        add_colorbar(fig, cax1, ax[0], [-1, 0, 1], "directed intensity ratio")
        add_colorbar(fig, cax2, ax[1], [0, 0.5, 1], "undirected intensity ratio")

        return ax, fig

    def plot_foi(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the field of interest of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        get_logger().info("Plotting: figure of interest")

        img = collection.get_image_by_img_name(img_name)

        im_junction = img.junction
        mask = img.segmentation.segmentation_mask_connected

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        single_cell_dataset = collection.dataset.loc[
            collection.dataset["filename"] == img_name
            ]
        foi_name = collection.get_runtime_params_by_img_name(
            img_name
        ).feature_of_interest
        foi = single_cell_dataset[foi_name]
        # figure and axes
        fig, ax = self._get_figure(1)
        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1)

        # plot the figure of interest
        m = np.copy(mask.data)
        for index, row in single_cell_dataset.iterrows():
            foi_val = row[foi_name]
            label = row["label"]

            m = np.where(mask.data == label, foi_val, m)

            ax.text(
                row["cell_X"],
                row["cell_Y"],
                str(np.round(row[foi_name], 1)),
                color=self.params.font_color,
                fontsize=self.params.fontsize_text_annotations,
            )

        cax = ax.imshow(np.ma.masked_where(m == 0, m), cmap=plt.cm.bwr, alpha=0.8)

        min = np.nanmin(foi)
        max = np.nanmax(foi)
        yticks = [min, np.round(min + (max - min) / 2, 1), max]
        add_colorbar(fig, cax, ax, yticks, foi_name)

        # set title and ax limits
        add_title(
            ax, "feature of interest", im_junction.data, self.params.show_graphics_axis
        )

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_foi",
            [ax],
            pixel_to_micron_ratio,
            close,
        )

        return fig, ax

    def plot_orientation(
            self, collection: PropertiesCollection, img_name: str, close: bool = False
    ):
        """Plots the orientation of a specific image in the collection

        Args:
            collection:
                The collection containing the features
            img_name:
                The name of the image to plot
            close:
                whether to close the figure after saving

        """
        img = collection.get_image_by_img_name(img_name)
        im_junction = img.junction
        instance_segmentation_con = img.segmentation.segmentation_mask_connected

        inst_nuclei_mask = None
        if img.has_nuclei():
            inst_nuclei_mask = img.nucleus.get_mask_by_name("nuclei_mask_seg")

        pixel_to_micron_ratio = img.img_params.pixel_to_micron_ratio

        get_logger().info("Plotting: orientation")

        # figure and axes
        number_sub_figs = 1
        if inst_nuclei_mask is not None:
            number_sub_figs = 2

        fig, ax = self._get_figure(number_sub_figs)

        # get cell_orientation
        cell_orientation_vec = collection.get_properties_by_img_name(img_name)[
            "cell_shape_orientation_deg"
        ].values
        cell_orientation = instance_segmentation_con.relabel(cell_orientation_vec)

        # add cell (and nuclei) orientation to the figure
        if inst_nuclei_mask is not None:
            Plotter._add_cell_orientation(fig, ax[0], im_junction, cell_orientation)

            # get nuclei orientation
            nuc_shape_orientation_rad_vector = collection.get_properties_by_img_name(
                img_name
            )["nuc_shape_orientation_rad"].values
            nuc_shape_orientation_deg_vector = (
                    nuc_shape_orientation_rad_vector * 180.0 / np.pi
            )
            nuclei_orientation = inst_nuclei_mask.relabel(
                nuc_shape_orientation_deg_vector
            )

            get_logger().info(
                "Maximal nuclei orientation: %s" % str(np.max(nuclei_orientation.data))
            )
            get_logger().info(
                "Minimal nuclei orientation: %s" % str(np.min(nuclei_orientation.data))
            )

            Plotter._add_nuclei_orientation(fig, ax[1], im_junction, nuclei_orientation)
        else:
            Plotter._add_cell_orientation(fig, ax, im_junction, cell_orientation)

        # plot major and minor axis
        for index, row in collection.get_properties_by_img_name(img_name).iterrows():
            if inst_nuclei_mask is not None:
                # plot orientation degree
                Plotter._add_single_cell_orientation_degree_axis(
                    ax[0],
                    row["cell_X"],
                    row["cell_Y"],
                    row["cell_shape_orientation_rad"],
                    row["cell_major_axis_length"],
                    row["cell_minor_axis_length"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )

                # plot orientation degree nucleus
                Plotter._add_single_cell_orientation_degree_axis(
                    ax[1],
                    row["nuc_X"],
                    row["nuc_Y"],
                    row["nuc_shape_orientation_rad"],
                    row["nuc_major_axis_length"],
                    row["nuc_minor_axis_length"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )
            else:
                # plot orientation degree
                Plotter._add_single_cell_orientation_degree_axis(
                    ax,
                    row["cell_X"],
                    row["cell_Y"],
                    row["cell_shape_orientation_rad"],
                    row["cell_major_axis_length"],
                    row["cell_minor_axis_length"],
                    self.params.fontsize_text_annotations,
                    self.params.font_color,
                    self.params.marker_size,
                )

        # set title and ax limits
        if inst_nuclei_mask is not None:
            add_title(
                ax[0],
                "cell shape orientation",
                im_junction.data,
                self.params.show_graphics_axis,
            )
            add_title(
                ax[1],
                "nuclei shape orientation",
                im_junction.data,
                self.params.show_graphics_axis,
            )
            axes = [ax[0], ax[1]]
        else:
            add_title(
                ax,
                "cell shape orientation",
                im_junction.data,
                self.params.show_graphics_axis,
            )
            axes = [ax]

        self._finish_plot(
            fig,
            collection.get_out_path_by_name(img_name),
            img_name,
            "_shape_orientation",
            axes,
            pixel_to_micron_ratio,
            close,
        )

        return fig, ax

    def _finish_plot(
            self,
            fig,
            output_path,
            img_name,
            output_suffix,
            axes,
            pixel_to_micron_ratio,
            close=False,
            image=None,
    ):
        # plot scale bar for this figure
        if self.params.plot_scalebar:
            for ax in axes:
                add_scalebar(
                    ax,
                    self.params.length_scalebar_microns,
                    pixel_to_micron_ratio,
                    int(self.params.length_scalebar_microns / 2),
                    self.params.font_color,
                )

        # save output & close
        save_current_fig(
            self.params.graphics_output_format,
            output_path,
            img_name,
            output_suffix,
            image=image,
        )

        # close figure
        if close:
            plt.close(fig)

    def plot_collection(self, collection: PropertiesCollection, close: bool = False):
        """Plot all features of all images in the collection

        Args:
            collection:
                the collection to plot
            close:
                whether to close the figure after saving

        """
        for key in collection.img_dict.keys():
            img = collection.get_image_by_img_name(key)
            r_params = collection.get_runtime_params_by_img_name(key)

            if self.params.plot_polarity and img.has_nuclei() and img.has_organelle():
                self.plot_organelle_polarity(collection, key, close)
                if img.has_nuclei():
                    self.plot_nuc_displacement_orientation(collection, key, close)

            if self.params.plot_marker and img.has_marker():
                self.plot_marker_expression(collection, key, close)
                self.plot_marker_polarity(collection, key, close)
                if img.has_nuclei():
                    self.plot_marker_nucleus_orientation(collection, key, close)

                if self.params.plot_ratio_method:
                    self.plot_marker_cue_intensity_ratio(collection, key, close)

            if self.params.plot_junctions and img.has_junction():
                self.plot_junction_polarity(collection, key, close)
                self.plot_corners(collection, key, close)

            if self.params.plot_orientation:
                self.plot_eccentricity(collection, key, close)

            if self.params.plot_ratio_method:
                self.plot_junction_cue_intensity_ratio(collection, key, close)

            if self.params.plot_cyclic_orientation:
                self.plot_orientation(collection, key, close)

            if self.params.plot_foi:
                if r_params.extract_group_features:
                    self.plot_foi(collection, key, close)

    @staticmethod
    def _add_single_cell_orientation_degree_axis(
            ax,
            y0,
            x0,
            orientation,
            major_axis_length,
            minor_axis_length,
            fontsize=3,
            font_color="w",
            markersize=2,
    ):
        (
            x1_ma,
            x1_mi,
            x2_ma,
            x2_mi,
            y1_ma,
            y1_mi,
            y2_ma,
            y2_mi,
        ) = Plotter._calc_single_cell_axis_orientation_vector(
            x0, y0, orientation, major_axis_length, minor_axis_length
        )
        orientation_degree = 180.0 * orientation / np.pi

        ax.plot((y1_ma, y2_ma), (x1_ma, x2_ma), "--w", linewidth=0.5)
        ax.plot((y1_mi, y2_mi), (x1_mi, x2_mi), "--w", linewidth=0.5)
        ax.plot(y0, x0, ".b", markersize=markersize)
        ax.text(
            y0,
            x0,
            str(int(np.round(orientation_degree, 0))),
            color=font_color,
            fontsize=fontsize,
        )

    @staticmethod
    def _add_nuclei_orientation(
            fig,
            ax,
            im_junction: BioMedicalChannel,
            nuclei_orientation: BioMedicalInstanceSegmentationMask,
    ):
        v_min = 0.0
        v_max = 180.0
        yticks = [0.0, 45.0, 90.0, 135.0, 180.0]

        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        # show nuclei orientation everywhere but background label

        cax_1 = ax.imshow(
            nuclei_orientation.mask_background().data,
            cmap=cm.cm.phase,
            vmin=v_min,
            vmax=v_max,
            alpha=0.75,
        )

        # colorbar
        add_colorbar(fig, cax_1, ax, yticks, "shape orientation (degree)")

    @staticmethod
    def _add_cell_orientation(
            fig,
            ax,
            im_junction: BioMedicalChannel,
            cell_orientation: BioMedicalInstanceSegmentationMask,
    ):
        v_min = 0.0
        v_max = 180.0
        yticks = [0.0, 45.0, 90.0, 135.0, 180.0]

        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        # show cell_orientation everywhere but background label
        cax = ax.imshow(
            cell_orientation.mask_background().data,
            cmap=cm.cm.phase,
            vmin=v_min,
            vmax=v_max,
            alpha=0.75,
        )

        # colorbar
        add_colorbar(fig, cax, ax, yticks, "shape orientation (degree)")

    @staticmethod
    def _calc_single_cell_axis_orientation_vector(
            x, y, orientation, major_axis_length, minor_axis_length
    ):
        x1_major = x + math.sin(orientation) * 0.5 * major_axis_length
        y1_major = y - math.cos(orientation) * 0.5 * major_axis_length
        x2_major = x - math.sin(orientation) * 0.5 * major_axis_length
        y2_major = y + math.cos(orientation) * 0.5 * major_axis_length

        x1_minor = x - math.cos(orientation) * 0.5 * minor_axis_length
        y1_minor = y - math.sin(orientation) * 0.5 * minor_axis_length
        x2_minor = x + math.cos(orientation) * 0.5 * minor_axis_length
        y2_minor = y + math.sin(orientation) * 0.5 * minor_axis_length

        return [
            x1_major,
            x1_minor,
            x2_major,
            x2_minor,
            y1_major,
            y1_minor,
            y2_major,
            y2_minor,
        ]

    @staticmethod
    def _add_nuclei_eccentricity(
            fig,
            ax,
            im_junction: BioMedicalChannel,
            nuclei_eccentricity: BioMedicalInstanceSegmentationMask,
    ):
        v_min = 0.0
        v_max = 1.0
        yticks = [0.0, 0.5, 1.0]

        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        # show nuclei eccentricity everywhere but background label
        cax_1 = ax.imshow(
            nuclei_eccentricity.mask_background().data,
            cmap=plt.cm.bwr,
            vmin=v_min,
            vmax=v_max,
            alpha=0.5,
        )

        # colorbar
        add_colorbar(fig, cax_1, ax, yticks, "eccentricity")

    @staticmethod
    def _add_cell_eccentricity(
            fig,
            ax,
            im_junction: BioMedicalChannel,
            cell_eccentricity: BioMedicalInstanceSegmentationMask,
    ):
        v_min = 0.0
        v_max = 1.0
        yticks = [0.0, 0.5, 1.0]

        ax.imshow(im_junction.data, cmap=plt.cm.gray, alpha=1.0)

        # show cell_eccentricity everywhere but background label
        cax_0 = ax.imshow(
            cell_eccentricity.mask_background().data,
            cmap=plt.cm.bwr,
            vmin=v_min,
            vmax=v_max,
            alpha=0.5,
        )

        # colorbar
        add_colorbar(fig, cax_0, ax, yticks, "eccentricity")

    @staticmethod
    def _add_single_cell_eccentricity_axis(
            ax,
            y0,
            x0,
            orientation,
            major_axis_length,
            minor_axis_length,
            eccentricity,
            fontsize=3,
            font_color="w",
            markersize=2,
    ):
        (
            x1_ma,
            x1_mi,
            x2_ma,
            x2_mi,
            y1_ma,
            y1_mi,
            y2_ma,
            y2_mi,
        ) = Plotter._calc_single_cell_axis_orientation_vector(
            x0, y0, orientation, major_axis_length, minor_axis_length
        )

        ax.plot((y1_ma, y2_ma), (x1_ma, x2_ma), "--w", linewidth=0.5)
        ax.plot((y1_mi, y2_mi), (x1_mi, x2_mi), "--w", linewidth=0.5)
        ax.plot(y0, x0, ".b", markersize=markersize)
        ax.text(
            y0, x0, str(np.round(eccentricity, 2)), color=font_color, fontsize=fontsize
        )



import math
from typing import List, Tuple

import numpy as np

from polarityjam.utils.decorators import experimental


def compute_reference_target_orientation_rad(ref_x: float, ref_y: float, target_x: float, target_y: float) -> float:
    """Computes the 2D orientation in rad between reference and target.

    Args:
        ref_x:
            The x coordinate of the reference point
        ref_y:
            The y coordinate of the reference point
        target_x:
            The x coordinate of the target point
        target_y:
            The y coordinate of the target point

    Returns:
        The orientation in rad between reference and target

    """
    vec_x = ref_x - target_x
    vec_y = ref_y - target_y
    organelle_orientation_rad = np.pi - np.arctan2(vec_x, vec_y)

    return organelle_orientation_rad


def compute_ref_x_abs_angle_deg(ref_x: float, ref_y: float, x: float, y: float) -> float:
    """Computes the angle between the x-axis and the vector (x,y).

    Args:
        ref_y:
            The y coordinate of the reference point
        ref_x:
            The x coordinate of the reference point
        x:
            The x coordinate
        y:
            The y coordinate

    Returns:
        The angle in rad

    """
    y_vec = y - ref_y
    x_vec = x - ref_x

    angle_deg = compute_angle_deg(np.arctan2(y_vec, x_vec))

    if angle_deg < 0:
        angle_deg = (180 - abs(angle_deg)) + 180

    return angle_deg


def compute_angle_deg(angle_rad: float) -> float:
    """Computes the angle given in rad in degrees.

    Args:
        angle_rad:
            The angle in rad

    Returns:
        The angle in degrees

    """
    return 180.0 * angle_rad / np.pi


def compute_shape_orientation_rad(orientation: float) -> float:
    """Computes the shape orientation (zero based) on x-axis.

    Args:
        orientation:
            The orientation

    Returns:
        The shape orientation in rad

    """
    # note, the values of orientation from props are in [-pi/2,pi/2] with zero along the y-axis
    return np.pi / 2.0 + orientation


def compute_marker_vector_norm(cell_x: float, cell_y: float, marker_centroid_x: float,
                               marker_centroid_y: float) -> float:
    """Computes the marker vector norm.

    Args:
        cell_x:
            The x coordinate of the cell
        cell_y:
            The y coordinate of the cell
        marker_centroid_x:
            The x coordinate of the marker centroid
        marker_centroid_y:
            The y coordinate of the marker centroid

    Returns:
        The marker vector norm

    """
    distance2 = (cell_x - marker_centroid_x) ** 2
    distance2 += (cell_y - marker_centroid_y) ** 2

    return np.sqrt(distance2)


@experimental
def map_single_cell_to_circle(sc_protein_area, x_centroid, y_centroid, r):
    """Maps a single cell to a circle. NOTE: EXPERIMENTAL"""

    circular_img = np.zeros([sc_protein_area.shape[0], sc_protein_area.shape[1]])
    circular_img_count = {}

    x_nonzero, y_nonzero = np.nonzero(sc_protein_area)

    # loop over bounding box indices
    for x, y in zip(x_nonzero, y_nonzero):
        x_vec = x_centroid - x
        y_vec = y_centroid - y

        angle_rad = np.pi - np.arctan2(x_vec, y_vec)
        new_x = r * np.sin(angle_rad) + x_centroid
        new_y = r * np.cos(angle_rad) + y_centroid

        # correct for wrong x axis alignment (bottom left corner is (0,0), not top left)
        new_x = x_centroid - (new_x - x_centroid)
        new_x = int(new_x)
        new_y = int(new_y)

        # count, TODO: check why new_x and new_y are sometimes out of the image boundaries
        if (new_x in range(0, circular_img.shape[0])) and (new_y in range(0, circular_img.shape[1])):
            circular_img[int(new_x), int(new_y)] += sc_protein_area[x, y]
            if (int(new_x), int(new_y)) not in circular_img_count.keys():
                circular_img_count[int(new_x), int(new_y)] = 1
            else:
                circular_img_count[int(new_x), int(new_y)] += 1
        # else:
        #    print("Circular image coords out of bounds")
        #    print("x: ", int(new_x), ", y:", int(new_y))
    #        if (int(new_x), int(new_y)) not in circular_img_count.keys():
    #            circular_img_count[int(new_x), int(new_y)] = 1
    #        else:
    #            circular_img_count[int(new_x), int(new_y)] += 1
    #
    #        circular_img[int(new_x), int(new_y)] += sc_protein_area[x, y]

    # calculate mean
    for k in circular_img_count.keys():
        circular_img[k[0], k[1]] = circular_img[k[0], k[1]] / circular_img_count[k]

    return circular_img


def straight_line_length(corners: List[Tuple[int, int]]) -> float:
    """Computes length between corners. Corner point assumed to be in the correct order.

    Args:
        corners:
            The corners of the cell in the correct order

    Returns:
        The straight line length between the consecutive corner points
    """
    dist = []
    for idx, c in enumerate(corners):
        x = c
        y = corners[idx + 1] if idx < (len(corners) - 1) else corners[0]

        # dist between the corner to the next
        dist.append(math.sqrt((x[0] - y[0]) ** 2 + (x[1] - y[1]) ** 2))

    return sum(dist)



import argparse
import sys
from typing import Callable

import polarityjam
from polarityjam.commandline import run, run_stack, run_key
from polarityjam.polarityjam_logging import configure_logger, get_logger, get_log_file, close_logger
from polarityjam.utils.io import create_path_recursively


def startup():
    """Entry points of `polarityjam`."""
    parser = create_parser()
    args = parser.parse_args()

    __run_subcommand(args, parser)


def __run_subcommand(args, parser):
    """Calls a specific subcommand."""
    command = ""
    try:
        command = sys.argv[1]  # command always expected at second position
    except IndexError:
        parser.error("Please provide a valid action!")

    create_path_recursively(args.out_path)

    log_file = get_log_file(args.out_path)
    configure_logger('INFO', logfile_name=log_file)

    get_logger().debug("Running %s subcommand..." % command)

    get_logger().info("Polarityjam Version %s" % polarityjam.__version__)

    args.func(args)  # execute entry point function

    close_logger()


def create_parser():
    """Creates the parser for the command line interface."""
    parser = PolarityjamParser()

    # run action
    p = parser.create_file_command_parser('run', run, 'Feature extraction from a single tiff image.')
    p.add_argument('in_file', type=str, help='Path to the input tif file.')
    p.add_argument('out_path', type=str, help='Path to the output folder.')
    p.add_argument('--filename_prefix', type=str, help='prefix for the output file.', required=False, default=None)

    # stack action
    p = parser.create_file_command_parser(
        'run-stack', run_stack, 'Feature extraction from an input folder containing several tiff files.'
    )
    p.add_argument('in_path', type=str, help='Name for the input folder containing tif images.')
    p.add_argument('out_path', type=str, help='Path to the output folder.')

    # key action
    p = parser.create_file_command_parser(
        'run-key',
        run_key,
        'Feature extraction from a given list of folders provided within a csv file.'
    )
    p.add_argument('in_path', type=str, help='Path prefix. Note: Base path for all folders listed in the key file.')
    p.add_argument(
        'in_key', type=str, help='Path to the input key file. File must contain column '
                                 '"folder_name" and "condition_name". Note: Folder name relative to the path prefix.'
    )
    p.add_argument('out_path', type=str, help='Path to the output folder.')

    return parser.parser


class ArgumentParser(argparse.ArgumentParser):
    """Override default error method of all parsers to show help of sub-command."""

    def error(self, message: str):
        self.print_help()
        self.exit(2, '%s: error: %s\n' % (self.prog, message))


class PolarityjamParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.parent_parser = self.create_parent_parser()
        self.parser = self.create_parser()
        self.subparsers = self.parser.add_subparsers(title='actions', help='sub-command help')

    @staticmethod
    def create_parent_parser() -> ArgumentParser:
        """Parent parser for all subparsers to have the same set of arguments."""
        parent_parser = ArgumentParser(add_help=False)
        parent_parser.add_argument('--version', '-V', action='version', version="%s " % polarityjam.__version__)
        # parse logging
        parent_parser.add_argument(
            '--log',
            required=False,
            help='Logging level for your command. Choose between %s' %
                 ", ".join(['INFO', 'DEBUG']),
            default='INFO'
        )
        return parent_parser

    def create_parser(self) -> ArgumentParser:
        """Creates the main parser for the framework."""
        parser = ArgumentParser(
            add_help=True,
            description='Polarityjam - feature extraction pipeline for multichannel images.',
            parents=[self.parent_parser])
        return parser

    def create_command_parser(self, command_name: str, command_function: Callable, command_help: str) -> ArgumentParser:
        """Creates a parser for a Polarityjam command, specified by a name, a function and a help description."""
        parser = self.subparsers.add_parser(command_name, help=command_help, parents=[self.parent_parser])
        parser.set_defaults(func=command_function)
        return parser

    def create_file_command_parser(self, command_name: str, command_function: Callable,
                                   command_help: str) -> ArgumentParser:
        """Creates a parser for a Polarityjam command dealing with a file.

        Parser is specified by a name, a function and a help description.
        """
        parser = self.create_command_parser(command_name, command_function, command_help)
        parser.add_argument('param', type=str, help='Path to the parameter file.')
        return parser



from __future__ import annotations

from hashlib import sha1
from typing import Dict, Union

import numpy as np

from polarityjam.model.parameter import ImageParameter
from polarityjam.model.masks import BioMedicalInstanceSegmentation, BioMedicalMask
from polarityjam.polarityjam_logging import get_logger


class BioMedicalChannel:  # todo: make it a PIL image for enhanced compatability?
    """Class representing a single channel of a biomedical image. It can contain multiple masks."""
    def __init__(self, channel: np.ndarray):
        self.data = channel
        self.masks: Dict[str, Union[BioMedicalMask, BioMedicalInstanceSegmentation]] = {}

    def mask(self, mask: BioMedicalMask) -> BioMedicalChannel:
        """Masks the channel with a given mask."""
        return BioMedicalChannel(self.data * mask.data)

    def add_mask(self, key: str, val: BioMedicalMask):
        """Adds a mask given a name to the channel."""
        self.masks[key] = val

    def remove_mask(self, key: str):
        """Removes a mask from the channel."""
        del self.masks[key]

    def get_mask_by_name(self, name: str) -> Union[BioMedicalMask, BioMedicalInstanceSegmentation]:
        """Returns a mask by its name."""
        return self.masks[name]


class BioMedicalImage:
    """Class representing a biomedical image. It can contain multiple channels and a segmentation."""
    def __init__(self, img: np.ndarray, img_params: ImageParameter,
                 segmentation: BioMedicalInstanceSegmentation = None):
        self.img = img
        self.segmentation = segmentation
        self.img_params = img_params
        self.marker = self.get_image_marker(img)
        self.junction = self.get_image_junction(img)
        self.nucleus = self.get_image_nucleus(img)
        self.organelle = self.get_image_organelle(img)
        self.img_hash = self.get_image_hash(img)

    def get_image_marker(self, img: np.ndarray):
        """Gets the image of the marker channel specified in the img_params.

        Args:
            img:
                The image to get the marker channel from.

        Returns:
            The np.ndarray of the marker channel.

        """
        if self.img_params.channel_expression_marker >= 0:
            get_logger().info("Marker channel at position: %s" % str(self.img_params.channel_expression_marker))
            return BioMedicalChannel(img[:, :, self.img_params.channel_expression_marker])
        return None

    def get_image_junction(self, img: np.ndarray):
        """Gets the image of the junction channel specified in the img_params.

        Args:
            img:
                The image to get the junction channel from.

        Returns:
            The np.ndarray of the junction channel.

        """
        if self.img_params.channel_junction >= 0:
            get_logger().info("Junction channel at position: %s" % str(self.img_params.channel_junction))
            return BioMedicalChannel(img[:, :, self.img_params.channel_junction])
        return None

    def get_image_nucleus(self, img: np.ndarray):
        """Gets the image of the nucleus channel specified in the img_params.

        Args:
            img:
                The image to get the nucleus channel from.

        Returns:
            The np.ndarray of the nucleus channel.

        """
        if self.img_params.channel_nucleus >= 0:
            get_logger().info("Nucleus channel at position: %s" % str(self.img_params.channel_nucleus))
            return BioMedicalChannel(img[:, :, self.img_params.channel_nucleus])
        return None

    def get_image_organelle(self, img: np.ndarray):
        """Gets the image of the organelle channel specified in the img_params.

        Args:
            img:
                The image to get the organelle channel from.

        Returns:
            The np.ndarray of the organelle channel.
        """
        if self.img_params.channel_organelle >= 0:
            get_logger().info("Organelle channel at position: %s" % str(self.img_params.channel_organelle))
            return BioMedicalChannel(img[:, :, self.img_params.channel_organelle])
        return None

    @staticmethod
    def get_image_hash(img: np.ndarray) -> str:
        """Returns the hash of the given image.

        Args:
            img:
                The image to get the hash from.

        Returns:
            The hash of the image.

        """
        return sha1(img.copy(order='C')).hexdigest()

    def has_nuclei(self) -> bool:
        """Returns whether the image has a nucleus channel."""
        return self.nucleus is not None

    def has_organelle(self) -> bool:
        """Returns whether the image has an organelle channel."""
        return self.organelle is not None

    def has_junction(self) -> bool:
        """Returns whether the image has a junction channel."""
        return self.junction is not None

    def has_marker(self) -> bool:
        """Returns whether the image has a marker channel."""
        return self.marker is not None



import os
from pathlib import Path

from polarityjam.utils.io import read_parameters


class Parameter:
    """Base class for all parameters."""

    def __init__(self, *args, **kwargs):
        # init with default values from resources
        current_path = Path(os.path.dirname(os.path.realpath(__file__)))
        param_base_file = Path(current_path).joinpath("..", "utils", "resources", "parameters.yml")
        args_init = read_parameters(str(param_base_file))

        for key in args_init:
            self._setattr(key, args_init[key])

        if args != ():
            for dictionary in args:
                for key in dictionary:
                    self._setattr(key, dictionary[key])

        if kwargs != {}:
            for key in kwargs:
                self._setattr(key, kwargs[key])

    def _setattr(self, key, val):
        if hasattr(self, key):
            setattr(self, key, val)

    def reset(self):
        for key in self.__dict__:
            self._setattr(key, None)

    @classmethod
    def from_yml(cls, path: str):
        params = read_parameters(path)

        cls(**params)

    def __str__(self, indent=1):
        s = '%s:  \n' % self.__class__.__name__
        for attr in self.__dict__:
            s += '{:<30}{:<40}\n'.format(attr, str(getattr(self, attr)))
        return s


class SegmentationParameter(Parameter):
    """Parameter class for the parameters necessary for the segmentation of images."""

    def __init__(self, attrs=None):
        if attrs is None:
            attrs = {}
        self.manually_annotated_mask = None
        self.store_segmentation = None
        self.use_given_mask = None
        self.model_type = None
        self.model_path = None
        self.estimated_cell_diameter = None
        self.flow_threshold = None
        self.cellprob_threshold = None
        self.use_gpu = None
        self.clear_border = None
        self.min_cell_size = None

        super().__init__(**attrs)


class ImageParameter(Parameter):
    """Parameter class for the parameters necessary for the image processing."""
    def __init__(self, attrs=None):
        if attrs is None:
            attrs = {}
        self.channel_junction = None
        self.channel_nucleus = None
        self.channel_organelle = None
        self.channel_expression_marker = None
        self.pixel_to_micron_ratio = None

        super().__init__(**attrs)

    def __len__(self):
        c = 0
        for key in self.__dict__:
            if getattr(self, key) is not None:
                c += 1

        return c


class RuntimeParameter(Parameter):
    """Parameter class for the parameters necessary for the calculation of the features."""
    def __init__(self, attrs=None):
        if attrs is None:
            attrs = {}
        self.extract_group_features = None
        self.membrane_thickness = None
        self.feature_of_interest = None
        self.min_cell_size = None
        self.min_nucleus_size = None
        self.min_organelle_size = None
        self.dp_epsilon = None
        self.cue_direction = None

        super().__init__(**attrs)


class PlotParameter(Parameter):
    """Parameter class for the parameters necessary for the plotting of the results."""
    def __init__(self, attrs=None):
        if attrs is None:
            attrs = {}
        self.plot_junctions = None
        self.plot_polarity = None
        self.plot_orientation = None
        self.plot_marker = None
        self.plot_ratio_method = None
        self.plot_cyclic_orientation = None
        self.plot_foi = None

        self.outline_width = None
        self.show_polarity_angles = None
        self.show_graphics_axis = None
        self.pixel_to_micron_ratio = None
        self.plot_scalebar = None
        self.length_scalebar_microns = None

        self.graphics_output_format = None
        self.dpi = None
        self.graphics_width = None
        self.graphics_height = None
        self.membrane_thickness = None

        self.fontsize_text_annotations = None
        self.font_color = None
        self.marker_size = None

        super().__init__(**attrs)



from typing import List, Set, Tuple

import numpy as np
from skimage.future.graph import RAG

from polarityjam.model.properties import NeighborhoodProps
from polarityjam.polarityjam_logging import get_logger


def shared_edges(rag: RAG, node: int) -> List[Set[int]]:
    """Return all first nearest neighbor nodes of a given node in a rag

    Args:
        rag:
            The RegionAdjacencyGraph.
        node:
            The node of interest.

    Returns:

    """
    first_nearest = list(set(rag.neighbors(node)))
    return first_nearest


def n_neighbors(rag: RAG, node: int) -> Tuple[List[Set[int]], List[int]]:
    """Get first and second nearest neighbors of a node in a rag

    Args:
        rag:
            The RegionAdjacencyGraph.
        node:
            The node of interest.

    Returns:
        Tuple of the first and second nearest neighbors of a node in a rag.

    """
    first_nearest_list = shared_edges(rag, node)

    # get list of second nearest neighbors
    k_nearest_list_unfiltered = []
    for first_nearest in first_nearest_list:
        k_nearest = list(rag.neighbors(first_nearest))
        k_nearest_list_unfiltered += k_nearest

    first_nearest_set = set(first_nearest_list)
    k_nearest_set = set(k_nearest_list_unfiltered)

    # get the difference of both sets and remove the node of interest from the final set
    second_nearest_list = list(k_nearest_set.difference(first_nearest_set))
    second_nearest_list = list(filter(lambda n: n != node, second_nearest_list))

    return first_nearest_list, second_nearest_list


def k_neighbor_dif(graph: RAG, foi: str) -> List[NeighborhoodProps]:
    """Extracts neighborhood statics from graph."

    Args:
        graph:
            The RegionAdjacencyGraph.
        foi:
            The feature of interest.

    Returns:
        List of NeighborhoodProps objects.

    """
    get_logger().info("Calculating first and second nearest neighbor statistic...")
    neighborhood_props_list = []
    for node in graph.nodes():
        neighborhood_props = NeighborhoodProps()
        neighbours = len(list(graph.neighbors(int(node))))

        neighborhood_props.num_neighbours = neighbours

        # feature of interest, first and second_nearest_neighbors
        node_foi = graph.nodes[node][foi]
        first_nearest_neighbors, second_nearest_neighbors = n_neighbors(graph, node)

        # feature of interest in comparison with first_nearest neighbor nodes
        foi_first_nearest_diff = [graph.nodes[neighbor][foi] - node_foi for neighbor in first_nearest_neighbors]

        # feature of interest in comparison to second nearest neighbor nodes
        foi_second_nearest_diff = [graph.nodes[neighbor][foi] - node_foi for neighbor in second_nearest_neighbors]

        # append first nearest neighbors statistics
        if len(foi_first_nearest_diff) > 0:
            neighborhood_props.mean_dif_first_neighbors = np.mean(foi_first_nearest_diff)
            neighborhood_props.median_dif_first_neighbors = np.median(foi_first_nearest_diff)
            neighborhood_props.var_dif_first_neighbors = np.std(foi_first_nearest_diff)
            neighborhood_props.range_dif_first_neighbors = abs(
                np.min(foi_first_nearest_diff) - np.max(foi_first_nearest_diff))

        # append second nearest neighbors statistics
        if len(foi_second_nearest_diff) > 0:
            neighborhood_props.mean_dif_second_neighbors = np.mean(foi_second_nearest_diff)
            neighborhood_props.median_dif_second_neighbors = np.median(foi_second_nearest_diff)
            neighborhood_props.var_dif_second_neighbors = np.std(foi_second_nearest_diff)
            neighborhood_props.range_dif_second_neighbors = abs(
                np.min(foi_second_nearest_diff) - np.max(foi_second_nearest_diff))

        neighborhood_props_list.append(neighborhood_props)

    return neighborhood_props_list



import sys
from pathlib import Path

import numpy as np
import tifffile
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar


def save_current_fig(graphics_output_format, output_path, filename, filename_suffix, image=None):
    # prevent text outside figure area
    plt.tight_layout()

    filename = str(filename)
    filename_suffix = str(filename_suffix)

    if "pdf" in graphics_output_format:
        plt.savefig(str(Path(output_path).joinpath(filename + filename_suffix + ".pdf")))
    if "svg" in graphics_output_format:
        plt.savefig(str(Path(output_path).joinpath(filename + filename_suffix + ".svg")))
    if "png" in graphics_output_format:
        plt.savefig(str(Path(output_path).joinpath(filename + filename_suffix + ".png")))
    if "tif" in graphics_output_format and image is not None:
        tifffile.imwrite(str(Path(output_path).joinpath(filename + filename_suffix + ".tif")), image)


def add_vector(ax, x_pos_p1, y_pos_p1, x_pos_p2, y_pos_p2, markersize=2, font_color="w"):
    ax.plot(x_pos_p1, y_pos_p1, '.g', markersize=markersize)
    ax.plot(x_pos_p2, y_pos_p2, '.m', markersize=markersize)
    ax.arrow(
        x_pos_p1,
        y_pos_p1,
        x_pos_p2 - x_pos_p1,
        y_pos_p2 - y_pos_p1,
        color=font_color,
        width=4
    )


def add_colorbar(fig, cax, ax, yticks, label):
    color_bar = fig.colorbar(cax, ax=ax, shrink=0.3)
    color_bar.set_label(label)
    color_bar.ax.set_yticks(yticks)


def add_scalebar(ax, length_scalebar_microns, pixel_to_micron_ratio, size_vertical, font_color="w"):
    length_scalebar_pixels = length_scalebar_microns * pixel_to_micron_ratio

    if sys.getdefaultencoding() == 'utf-8':
        entity = "\u03BC"
    else:
        entity = "mu"
    text = "%s %sm" % (length_scalebar_microns, entity)

    scalebar = AnchoredSizeBar(
        ax.transData,
        length_scalebar_pixels,
        text,
        'lower right',
        pad=0.1,
        color=font_color,
        frameon=False,
        size_vertical=size_vertical,
    )
    ax.add_artist(scalebar)


def add_title(ax, plot_title: str, im_junction: np.ndarray, axis_on: bool):
    ax.set_title(plot_title)
    ax.set_xlim(0, im_junction.shape[1])  # x-axis is second dimension in numpy array
    ax.set_ylim(0, im_junction.shape[0])
    ax.invert_yaxis()
    if not axis_on:
        ax.axis('off')





import logging
import sys
from pathlib import Path

from polarityjam.utils.io import get_doc_file_prefix

LOGGER_NAME = 'polarityjam'


def get_logger():
    return logging.getLogger(LOGGER_NAME)  # root logger


def get_log_file(out_folder):
    log_file = Path(out_folder).joinpath("%s.log" % get_doc_file_prefix())
    log_file.touch()

    return log_file


def configure_logger(loglevel=None, logfile_name=None, formatter_string=None):
    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(loglevel)

    # create formatter
    if not formatter_string:
        formatter = get_default_formatter()
    else:
        formatter = logging.Formatter(formatter_string)

    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(loglevel)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if logfile_name:
        ch = logging.FileHandler(logfile_name, mode='a', encoding=None, delay=False)
        ch.setLevel(loglevel)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    return logger


def close_logger():
    for h in logging.getLogger(LOGGER_NAME).handlers:
        if isinstance(h, logging.FileHandler):
            h.close()
    logging.getLogger(LOGGER_NAME).handlers.clear()


def get_default_formatter():
    return logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%H:%M:%S')



import fileinput
import importlib.util
import os
import platform
import sys
import unittest
from pathlib import Path

from nbconvert import PythonExporter

from polarityjam.test.test_common import TestCommon


class TestIntegration(TestCommon):
    def setUp(self) -> None:
        super().setUp()
        if not self.data_path.exists():
            self.extract_test_data()

    def tearDown(self) -> None:
        super().tearDown()

    def nb_to_py(self, path_nb, filename, outpath_py, prefix_str="tmp_"):
        outpath_py = Path(outpath_py)

        # export
        exp = PythonExporter()
        nb_python_mem, _ = exp.from_filename(str(path_nb))

        p = outpath_py.joinpath(prefix_str + filename + ".py")
        with open(p, "w+") as f:
            f.write(nb_python_mem)

        return p

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_nb_2(self):
        # copy to notebook tmp dir
        path_nb = Path(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent.joinpath(
            "docs", "notebooks", "polarityjam-notebook_api.ipynb"
        )
        p = self.nb_to_py(path_nb, path_nb.stem, self.tmp_dir)

        # remove ipython formatting and set data path
        replace_cell_pattern = "### ADAPT ME ###"
        with fileinput.FileInput(p, inplace=True) as f:
            myiter = iter(f)
            for line in f:
                if line.startswith("get_ipython"):
                    print("")
                elif line.startswith("%"):
                    print("")
                elif line.startswith("!"):
                    print("")
                elif replace_cell_pattern:
                    if line.startswith(replace_cell_pattern):

                        print("input_file1 = Path(\"%s\")" % self.data_path.parent.joinpath(
                            "data", "golgi_nuclei", "set_2", "060721_EGM2_18dyn_02.tif"))
                        print("input_file2 = Path(\"%s\")" % self.data_path.parent.joinpath(
                            "data", "golgi_nuclei", "set_1", "060721_EGM2_18dyn_01.tif"))

                        print("output_path = Path(\"%s\")" % self.data_path.parent)
                        print("output_file_prefix1 = \"060721_EGM2_18dyn_02\"")
                        print("output_file_prefix2 = \"060721_EGM2_18dyn_01\"")

                        while not next(myiter, "None").startswith(replace_cell_pattern):
                            print("")
                    else:
                        print(line, end='')
                else:
                    print(line, end='')

        root = Path(p).parent
        sys.path.insert(0, str(root))
        spec = importlib.util.spec_from_file_location("nb_test2", p)
        nb_test2 = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(nb_test2)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_nb_3(self):
        # copy to notebook tmp dir
        path_nb = Path(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent.joinpath(
            "docs", "notebooks", "polarityjam-notebook_vis.ipynb"
        )
        p = self.nb_to_py(path_nb, path_nb.stem, self.tmp_dir)

        # remove ipython formatting and set data path
        replace_cell_pattern = "### ADAPT ME ###"
        with fileinput.FileInput(p, inplace=True) as f:
            myiter = iter(f)
            for line in f:
                if line.startswith("get_ipython"):
                    print("")
                elif line.startswith("%"):
                    print("")
                elif line.startswith("!"):
                    print("")
                elif replace_cell_pattern:
                    if line.startswith(replace_cell_pattern):
                        print("output_path = Path(\"%s\")" % self.data_path.parent)
                        print("input_file = Path(\"%s\")" % self.data_path.parent.joinpath(
                            "data", "golgi_nuclei", "set_2", "060721_EGM2_18dyn_02.tif"))
                        print("output_file_prefix = \"060721_EGM2_18dyn_02\"")
                        while not next(myiter, "None").startswith(replace_cell_pattern):
                            print("")
                    else:
                        print(line, end='')
                else:
                    print(line, end='')

        root = Path(p).parent
        sys.path.insert(0, str(root))
        spec = importlib.util.spec_from_file_location("nb_test3", p)
        nb_test3 = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(nb_test3)

    @unittest.skipIf(platform.system().lower() == 'windows', "Plotting too memory extensive. Skipping test!")
    def test_nb(self):
        # copy to notebook tmp dir
        path_nb = Path(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent.joinpath(
            "docs", "notebooks", "polarityjam-notebook.ipynb"
        )
        p = self.nb_to_py(path_nb, path_nb.stem, self.tmp_dir)

        # remove ipython formatting and set data path
        replace_cell_pattern = "### ADAPT ME ###"
        with fileinput.FileInput(p, inplace=True) as f:
            myiter = iter(f)
            for line in f:
                if line.startswith("get_ipython"):
                    print("")
                elif line.startswith("%"):
                    print("")
                elif line.startswith("!"):
                    print("")
                elif replace_cell_pattern:
                    if line.startswith(replace_cell_pattern):
                        print("output_path = Path(\"%s\")" % self.data_path.parent)
                        print("input_file = Path(\"%s\")" % self.data_path.parent.joinpath(
                            "data", "golgi_nuclei", "set_2", "060721_EGM2_18dyn_02.tif"))
                        print("output_file_prefix = \"060721_EGM2_18dyn_02\"")
                        while not next(myiter, "None").startswith(replace_cell_pattern):
                            print("")
                    else:
                        print(line, end='')
                else:
                    print(line, end='')

        root = Path(p).parent
        sys.path.insert(0, str(root))
        spec = importlib.util.spec_from_file_location("nb_test", p)
        nb_test = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(nb_test)



import warnings
import functools


def experimental(func):
    """This is a decorator which can be used to mark functions
    as experimental. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', UserWarning)  # turn off filter
        warnings.warn("Call to experimental function {}. Handle results with care!".format(func.__name__),
                      category=UserWarning,
                      stacklevel=2)
        warnings.simplefilter('default', UserWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func



from pathlib import Path

import networkx as nx
import numpy as np
import yaml

from polarityjam.model.moran import run_morans
from polarityjam.test.test_common import TestCommon
from polarityjam.utils.io import read_parameters, read_image


class TestFunctions(TestCommon):
    """Small tests to ensure basic IO functionality of the tool."""

    def setUp(self) -> None:
        super().setUp()

    def test_read_parameters(self):
        # prepare
        local_parameter_file = {
            "channel_junction": 0,
            "channel_nucleus": 1,
            "channel_organelle": 2
        }
        local_parameter_file_path = Path(self.tmp_dir).joinpath("local_parameter_file.yml")

        with open(local_parameter_file_path, "w+") as file:
            file.write(yaml.dump(local_parameter_file, Dumper=yaml.Dumper))

        self.parameters["channel_junction"] = 0
        self.parameters["channel_nucleus"] = 1
        self.parameters["channel_organelle"] = 2

        # call
        r = read_parameters(local_parameter_file_path)

        # assert
        self.assertDictEqual(self.parameters, r)

    def test_read_img(self):
        # prepare
        test_image_path = self.get_test_image_path("060721_EGM2_18dyn_01.tif")

        # call
        r = read_image(test_image_path)

        # assert
        self.assertIsNotNone(r)

        # assert correct type
        self.assertEqual(type(np.array(0)), type(r))

    def test_morans_i_anti_correlation(self):
        # prepare anti-correlation sample
        graph = nx.grid_2d_graph(10, 10)

        for i in range(0, 10):
            if (i % 2) == 0:
                for j in range(0, 10):
                    if (j % 2) == 0:
                        nx.set_node_attributes(graph, {(i, j): 0}, name="morani")
                    else:
                        nx.set_node_attributes(graph, {(i, j): 1}, name="morani")
            else:
                for j in range(0, 10):
                    if (j % 2) == 0:
                        nx.set_node_attributes(graph, {(i, j): 1}, name="morani")

                    else:
                        nx.set_node_attributes(graph, {(i, j): 0}, name="morani")

        # run morans i
        mr_i = run_morans(graph, "morani")

        # check the output
        self.assertEqual(-1, mr_i.I)

    def test_morans_i_random(self):
        # prepare random sample
        graph = nx.grid_2d_graph(10, 10)

        empty_int = np.zeros((10, 10))
        for i in range(0, 10):
            for j in range(0, 10):
                rand_int = np.random.normal(0, 1)
                if rand_int >= .5:
                    empty_int[i, j] = 1
                else:
                    empty_int[i, j] = 0
                nx.set_node_attributes(graph, {(i, j): rand_int}, name="morani")
        # run
        mr_i = run_morans(graph, "morani")

        # assert
        self.assertAlmostEqual(0, mr_i.I, delta=0.3)

    def test_morans_i_correlated(self):
        # prepare correlated sample
        graph = nx.grid_2d_graph(10, 10)

        empty_int = np.zeros((10, 10))
        for i in range(0, 10):
            for j in range(0, 10):
                rand_int = 1.1
                empty_int[i, j] = rand_int
                nx.set_node_attributes(graph, {(i, j): rand_int}, name="morani")

        # run
        mr_i = run_morans(graph, "morani")

        # assert
        self.assertEqual(1, mr_i.I)


class EmptyTestClass:
    pass



import os
import tempfile
import unittest
from pathlib import Path

import yaml
from pyunpack import Archive

import polarityjam.test.test_config as config
from polarityjam.polarityjam_logging import close_logger
from polarityjam.utils.io import list_files_recursively


class TestCommon(unittest.TestCase):
    """Base class all tests inherit from. Responsible for temporary directory, cleanup, parameters."""

    def setUp(self) -> None:
        if config._TARGET_DIR:
            self.tmp_dir_ = Path(config._TARGET_DIR)
            self.tmp_dir = self.tmp_dir_
            self.output_path = self.tmp_dir.joinpath("output")
            self.data_path = self.tmp_dir.joinpath("data")
        else:
            self.tmp_dir_ = tempfile.TemporaryDirectory()
            self.tmp_dir = self.tmp_dir_.name
            self.output_path = Path(self.tmp_dir).joinpath("output")
            self.data_path = Path(self.tmp_dir).joinpath("data")

        self.current_path = Path(os.path.dirname(os.path.realpath(__file__)))
        self.parameters, self.param_base_file = self.load_parameters()

    def tearDown(self) -> None:
        close_logger()
        if isinstance(self.tmp_dir_, tempfile.TemporaryDirectory):
            self.tmp_dir_.cleanup()

    def extract_test_data(self):
        target = str(self.tmp_dir)
        Archive(str(self.current_path.joinpath("resources", "data.zip"))).extractall(str(target))

    def get_test_image_path(self, image_name):
        if not self.data_path.exists():
            self.extract_test_data()

        files_list = list_files_recursively(self.data_path)
        test_image_path = files_list[[i.stem + i.suffix for i in files_list].index(image_name)]

        return test_image_path

    def get_test_image_folder(self, mode):
        # "gn" = golgi_nuclei, "g" = "no golgi", "n" = "no nuclei"

        if not self.data_path.exists():
            self.extract_test_data()

        if mode == "gn":
            return self.data_path.joinpath("golgi_nuclei")
        elif mode == "n":
            return self.data_path.joinpath("no_nuclei")
        elif mode == "g":
            return self.data_path.joinpath("no_golgi")
        else:
            return None

    def get_test_parameter_file(self, name):
        return self.current_path.joinpath("resources", name)

    def get_test_key_file(self):
        return self.current_path.joinpath("resources", "test_key_file.csv")

    def load_parameters(self):
        param_base_file = Path(self.current_path).joinpath("..", "utils", "resources", "parameters.yml")

        with open(param_base_file, 'r') as yml_f:
            parameters = yaml.safe_load(yml_f)

        parameters["channel_junction"] = 3
        parameters["channel_nucleus"] = 2
        parameters["channel_organelle"] = 0

        return parameters, param_base_file



import os
from pathlib import Path
from typing import Union

import numpy as np

from polarityjam import PropertiesCollection
from polarityjam.controller.collector import (GroupPropertyCollector,
                                              PropertyCollector,
                                              SingleCellMaskCollector,
                                              SingleCellPropertyCollector)
from polarityjam.model.image import BioMedicalImage
from polarityjam.model.masks import (BioMedicalInstanceSegmentation,
                                     BioMedicalInstanceSegmentationMask,
                                     BioMedicalMask, SingleCellMasksCollection)
from polarityjam.model.parameter import ImageParameter, RuntimeParameter
from polarityjam.polarityjam_logging import get_logger


class Extractor:
    def __init__(self, params: RuntimeParameter):
        self.params = params
        self.collector = PropertyCollector()

    def threshold_size(self, sc_masks: SingleCellMasksCollection) -> bool:
        """Thresholds a collection of single cell mask based on cell size, nucleus size and organelle size.

        Args:
            sc_masks:
                The collection of single cell masks to threshold.

        Returns:
            True if the size count was below the threshold, False otherwise.

        """
        if sc_masks.sc_mask.is_count_below_threshold(self.params.min_cell_size):
            return True
        if sc_masks.sc_nucleus_mask is not None:
            if sc_masks.sc_nucleus_mask.is_count_below_threshold(
                self.params.min_nucleus_size
            ):
                return True
        if sc_masks.sc_organelle_mask is not None:
            if sc_masks.sc_organelle_mask.is_count_below_threshold(
                self.params.min_organelle_size
            ):
                return True
        return False

    def extract_cell_features(
        self, collection, bio_med_image, bio_med_segmentation, filename_prefix
    ):
        """Extracts features from cells

        Args:
            collection:
                PropertiesCollection object to which the extracted features will be added.
            bio_med_image:
                BioMedicalImage object containing the image information.
            bio_med_segmentation:
                BioMedicalInstanceSegmentation object containing the segmentation of the cells.
            filename_prefix:
                Name prefix for the image used for all produced output.

        """
        nuclei_mask_seg = None
        if bio_med_image.has_nuclei():
            nuclei_mask_seg = BioMedicalMask.from_threshold_otsu(
                bio_med_image.nucleus.data
            ).overlay_instance_segmentation(
                bio_med_segmentation.segmentation_mask_connected
            )

        organelle_mask_seg = None
        if bio_med_image.has_organelle():
            organelle_mask_seg = BioMedicalMask.from_threshold_otsu(
                bio_med_image.organelle.data
            ).overlay_instance_segmentation(
                bio_med_segmentation.segmentation_mask_connected
            )

        excluded = 0
        sc_masks_list = []
        # iterate through each unique segmented cell
        for (
            connected_component_label
        ) in bio_med_segmentation.segmentation_mask_connected.get_labels():

            sc_masks = SingleCellMaskCollector.calc_sc_masks(
                bio_med_image,
                connected_component_label,
                self.params.membrane_thickness,
                nuclei_mask_seg,
                organelle_mask_seg,
            )

            # threshold
            if self.threshold_size(sc_masks):
                get_logger().info(
                    'Cell "%s" falls under threshold! Removed from analysis!'
                    % connected_component_label
                )
                # remove a cell from the segmentation
                removed_islands_labels = bio_med_segmentation.remove_instance_label(
                    connected_component_label
                )

                if removed_islands_labels:
                    get_logger().info(
                        'Cell(s) "%s" became isolated and have additionally ben removed from the analysis!'
                        % ", ".join(removed_islands_labels)
                    )

                # remove instance from the organelle instance mask
                if bio_med_image.has_organelle():
                    organelle_mask_seg = organelle_mask_seg.remove_instance(
                        connected_component_label
                    )
                    # additionally remove the islands from the organelle mask
                    [
                        organelle_mask_seg.remove_instance(i)
                        for i in removed_islands_labels
                    ]

                # remove instance from the nucleus instance mask
                if bio_med_image.has_nuclei():
                    nuclei_mask_seg = nuclei_mask_seg.remove_instance(
                        connected_component_label
                    )
                    # additionally remove the islands from the nuclei mask
                    [nuclei_mask_seg.remove_instance(i) for i in removed_islands_labels]

                # exclude all islands from the analysis - e.g. keep only those that have not been removed from the RAG
                sc_masks_list = [
                    i
                    for i in sc_masks_list
                    if i.connected_component_label not in removed_islands_labels
                ]

                # counter for excluded cells
                excluded += 1 + len(removed_islands_labels)
                continue

            sc_masks_list.append(sc_masks)

        if bio_med_image.has_organelle():
            bio_med_image.organelle.add_mask("organelle_mask_seg", organelle_mask_seg)

        if bio_med_image.has_nuclei():
            bio_med_image.nucleus.add_mask("nuclei_mask_seg", nuclei_mask_seg)

        num_cells = len(bio_med_segmentation.segmentation_mask_connected) - excluded
        get_logger().info("Excluded cells: %s" % str(excluded))
        get_logger().info("Leftover cells: %s" % str(num_cells))

        # calculate properties for each cell
        for sc_masks in sc_masks_list:
            sc_props_collection = SingleCellPropertyCollector.calc_sc_props(
                sc_masks, bio_med_image, self.params
            )

            PropertyCollector.collect_sc_props(
                sc_props_collection,
                collection,
                filename_prefix,
                bio_med_image.img_hash,
                sc_masks.connected_component_label,
            )

    def extract(
        self,
        img: np.ndarray,
        img_params: ImageParameter,
        segmentation_mask: np.ndarray,
        filename_prefix: str,
        output_path: Union[Path, str],
        collection: PropertiesCollection,
    ) -> PropertiesCollection:
        """Extracts features from an input image into a given collection.

        Args:
            img:
                np.ndarray of the image to be processed.
            img_params:
                ImageParameter object containing the image parameters.
            segmentation_mask:
                np.ndarray of the cells mask.
            filename_prefix:
                Name prefix for the image used for all produced output.
            output_path:
                Path to the output directory.
            collection:
                PropertiesCollection object to which the extracted features will be added.

        Returns:
            PropertiesCollection object containing the extracted features.

        """
        filename_prefix, _ = os.path.splitext(os.path.basename(filename_prefix))

        bio_med_segmentation_mask = BioMedicalInstanceSegmentationMask(
            segmentation_mask
        )
        bio_med_segmentation = BioMedicalInstanceSegmentation(bio_med_segmentation_mask)

        bio_med_image = BioMedicalImage(
            img, img_params, segmentation=bio_med_segmentation
        )

        self.extract_cell_features(
            collection, bio_med_image, bio_med_segmentation, filename_prefix
        )

        if self.params.extract_group_features:
            self.extract_group_features(
                collection, bio_med_segmentation, filename_prefix
            )

        # mark the beginning of a new image that is potentially extracted
        PropertyCollector.set_reset_index(collection)
        PropertyCollector.add_out_path(collection, filename_prefix, output_path)
        PropertyCollector.add_runtime_params(collection, filename_prefix, self.params)
        PropertyCollector.add_img(collection, filename_prefix, bio_med_image)

        get_logger().info("Done feature extraction for file: %s" % str(filename_prefix))

        return collection

    def extract_group_features(
        self,
        collection: PropertiesCollection,
        bio_med_segmentation: BioMedicalInstanceSegmentation,
        filename_prefix: str,
    ):
        """Extracts features from a group of cells.

        Args:
            bio_med_segmentation:
                BioMedicalInstanceSegmentation object containing the segmentation of the cells.
            collection:
                PropertiesCollection object to which the extracted features will be added.
            filename_prefix:
                Name prefix for the image used for all produced output.

        """

        if len(collection) < 2:
            get_logger().warn(
                """Neighborhood analysis not possible.
                 Not enough cells to calculate group features!
                 Switch off neighborhood analysis (extract_group_features = False) or improve segmentation!"""
            )
            return

        foi_vec = collection.get_properties_by_img_name(filename_prefix)[
            self.params.feature_of_interest
        ].values
        bio_med_segmentation.set_feature_of_interest(
            self.params.feature_of_interest, foi_vec
        )

        # morans I analysis based on FOI
        morans_i = GroupPropertyCollector.calc_moran(
            bio_med_segmentation, self.params.feature_of_interest
        )
        PropertyCollector.collect_group_statistic(
            collection, morans_i, len(bio_med_segmentation.segmentation_mask_connected)
        )

        # neighborhood feature analysis based on FOI
        neighborhood_props_list = GroupPropertyCollector.calc_neighborhood(
            bio_med_segmentation, self.params.feature_of_interest
        )
        PropertyCollector.collect_neighborhood_props(
            collection, neighborhood_props_list
        )



from pathlib import Path
from typing import Tuple, Union

import cellpose.models
import numpy as np
import skimage
from skimage import morphology

from polarityjam.model.parameter import SegmentationParameter, ImageParameter
from polarityjam.polarityjam_logging import get_logger

from abc import ABCMeta, abstractmethod


class Segmenter:
    """Abstract class for an object performing a segmentation procedure."""
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, params: SegmentationParameter):
        self.params = params

    @abstractmethod
    def segment(self, img: np.ndarray, path: str=None) -> np.ndarray:
        """Should perform segmentation and return a mask image. Path can point to a model to load/state index/parameter
        file or something else needed to load a checkpoint needed for segmentation.

        Args:
            img:
                The image prepared for segmentation.

            path:
                Path to a model to load/state index/parameter file or smth. else needed to load a checkpoint needed
                to perform the segmentation.

        Returns:
            A mask as np.ndarray image.

        """
        raise NotImplementedError

    @abstractmethod
    def prepare(self, img: np.ndarray, input_parameter: ImageParameter) -> Tuple[np.ndarray, ImageParameter]:
        """Should perform preparation for a given image to perform the segmentation. Should return prepared image
        and its parameters. This could be a resizing, cropping, selecting channels, etc. E.g. whatever is needed to
        perform the segmentation.

        Args:
            img:
                The input image to prepare for segmentation as a numpy array.
            input_parameter:
                The parameters of the input image

        Returns:
            A tuple of the prepared image and its parameters as ImageParameter object.

        """
        raise NotImplementedError


class CellposeSegmenter(Segmenter):
    """Cellpose segmentation class"""

    def __init__(self, params: SegmentationParameter):
        super().__init__(params)
        self.params = params

    def segment(self, img: np.ndarray, path: Union[Path, str]=None) -> np.ndarray:
        """Performs the segmentation of the given image. If a path is given, the model is loaded from the given path.

        Args:
            img:
                The image prepared for segmentation.
            path:
                Path to a model to load/state index/parameter file or smth. else needed to load a checkpoint needed

        Returns:
            A mask as np.ndarray image.

        """
        return self._load_or_get_cellpose_segmentation(img, path)

    def prepare(self, img: np.ndarray, img_parameter: ImageParameter) -> Tuple[np.ndarray, ImageParameter]:
        """Prepares the image for segmentation. Returns an image that has the junction channel first, then the nucleus
        channel and the last channel is the cytoplasm channel.

        Args:
            img:
                The input image to prepare for segmentation as a numpy array.
            img_parameter:
                The parameters of the input image

        Returns:
            A tuple of the prepared image and its parameters as ImageParameter object.

        """

        get_logger().info("Image shape: %s" % str(img.shape))

        params_prep_img = ImageParameter()
        px_to_m_r = img_parameter.pixel_to_micron_ratio
        params_prep_img.reset()
        params_prep_img.pixel_to_micron_ratio = px_to_m_r

        im_junction = None
        im_nucleus = None

        if img_parameter.channel_junction >= 0:
            get_logger().info("Junction channel at position: %s" % str(img_parameter.channel_junction))
            im_junction = img[:, :, img_parameter.channel_junction]
            params_prep_img.channel_junction = 0

        if img_parameter.channel_nucleus >= 0:
            get_logger().info("Nucleus channel at position: %s" % str(img_parameter.channel_nucleus))
            im_nucleus = img[:, :, img_parameter.channel_nucleus]
            params_prep_img.channel_nucleus = 1

        if im_nucleus is not None:
            return np.array([im_junction, im_nucleus]), params_prep_img
        else:
            return im_junction, params_prep_img

    def _get_cellpose_model(self):
        """Gets the specified cellpose model"""

        if self.params.model_type == "custom":
            get_logger().info("Loading custom model from: %s" % self.params.model_path)
            model = cellpose.models.CellposeModel(gpu=self.params.use_gpu, pretrained_model=self.params.model_path)
        else:
            model = cellpose.models.Cellpose(gpu=self.params.use_gpu, model_type=self.params.model_type)

        return model

    def _get_cellpose_segmentation(self, im_seg, filepath):
        """Gets the cellpose segmentation. Expects im_seg to have junction channel first, then nucleus channel."""
        get_logger().info("Calculate cellpose segmentation. This might take some time...")
        get_logger().info("Using model type '%s' with estimated cell diameter %s, cellprob_threshold %s and flow threshold %s" % (
        self.params.model_type, self.params.estimated_cell_diameter, self.params.cellprob_threshold, self.params.flow_threshold))
        model = self._get_cellpose_model()
        if im_seg.ndim > 1:
            channels = [1, 2]
        else:
            channels = [0, 0]

        # masks, flows, styles, diams = model.eval(im_seg, channels=channels)

        if self.params.model_type == "custom":
            masks, flows, styles = model.eval(im_seg, diameter=self.params.estimated_cell_diameter,
                                              flow_threshold=self.params.flow_threshold,
                                              cellprob_threshold=self.params.cellprob_threshold,
                                              channels=channels)
        else:
            masks, flows, styles, diams = model.eval(im_seg, diameter=self.params.estimated_cell_diameter,
                                                     flow_threshold=self.params.flow_threshold,
                                                     cellprob_threshold = self.params.cellprob_threshold,
                                                     channels=channels)

        if self.params.store_segmentation:
            segmentation_list = {"masks": masks}
            segmentation, _ = self._get_segmentation_file_name(filepath)

            get_logger().info("Storing cellpose segmentation: %s" % segmentation)
            np.save(str(segmentation), segmentation_list, allow_pickle=True)

        return masks

    def _get_segmentation_file_name(self, filepath):
        stem = Path(filepath).stem

        suffix = "_seg.npy"
        if self.params.manually_annotated_mask:
            suffix = self.params.manually_annotated_mask
        segmentation = Path(filepath).parent.joinpath(stem + suffix)

        return segmentation, stem

    def _load_or_get_cellpose_segmentation(self, img_seg, filepath):
        get_logger().info("Look up cellpose segmentation...")
        segmentation, _ = self._get_segmentation_file_name(filepath)

        if segmentation.exists() and self.params.use_given_mask:
            get_logger().info("Load existing segmentation from %s ..." % segmentation)

            # in case an annotated mask is available
            cellpose_seg = np.load(str(segmentation), allow_pickle=True)
            cellpose_mask = cellpose_seg.item()['masks']

        else:
            cellpose_mask = self._get_cellpose_segmentation(img_seg, filepath)

        if self.params.clear_border:
            cellpose_mask_clear_border = skimage.segmentation.clear_border(cellpose_mask)
            number_of_cellpose_borders = len(np.unique(cellpose_mask)) - len(np.unique(cellpose_mask_clear_border))
            cellpose_mask = cellpose_mask_clear_border

            get_logger().info("Removed number of cellpose borders: %s" % number_of_cellpose_borders)

            cellpose_mask_remove_small_objects = morphology.remove_small_objects(
                cellpose_mask, self.params.min_cell_size, connectivity=2
            )
            number_of_cellpose_small_objects = len(np.unique(cellpose_mask)) - len(
                np.unique(cellpose_mask_remove_small_objects))
            cellpose_mask = cellpose_mask_remove_small_objects

            get_logger().info("Removed number of small objects: %s" % number_of_cellpose_small_objects)

        get_logger().info("Detected number of cellpose labels: %s" % len(np.unique(cellpose_mask)))

        return cellpose_mask



__version__ = "0.1.0"
__author__ = "Jan Philipp Albrecht, Wolfgang Giese"

# imports for python API - do not delete!
from polarityjam.model.parameter import RuntimeParameter, PlotParameter, SegmentationParameter, ImageParameter
from polarityjam.model.collection import PropertiesCollection
from polarityjam.model.masks import BioMedicalMask, BioMedicalInstanceSegmentationMask, BioMedicalInstanceSegmentation
from polarityjam.model.image import BioMedicalImage
from polarityjam.controller.extractor import Extractor
from polarityjam.controller.plotter import Plotter
from polarityjam.controller.segmenter import CellposeSegmenter



from typing import Union, List, Tuple

import cv2
import numpy as np
from shapely.affinity import rotate
from shapely.geometry import LineString
from shapely.geometry.polygon import Polygon
from shapely.ops import split

from polarityjam.compute.compute import compute_ref_x_abs_angle_deg
from polarityjam.compute.corner import get_contour


def mask_from_contours(ref_img: np.ndarray, coord_list_x: np.ndarray, coord_list_y: np.ndarray) -> np.ndarray:
    """Create a mask from a list of coordinates.

    Args:
        ref_img:
            The reference image
        coord_list_x:
            The x coordinates
        coord_list_y:
            The y coordinates

    Returns:
        The mask

    """
    mask = np.zeros(ref_img.shape, dtype=np.uint8)

    l = []
    for a, b in zip(coord_list_x, coord_list_y):
        l.append([a, b])

    mask = cv2.drawContours(mask, [np.array(l).astype(np.int32)], -1, 1, thickness=cv2.FILLED)

    return mask


def partition_single_cell_mask(sc_mask: np.ndarray, cue_direction: int,
                               major_axes_length: Union[int, float], num_partitions: int) -> Tuple[
    List[np.ndarray], List[Polygon]]:
    """Partitions a single cell mask into multiple masks from its centroid.

    Args:
        sc_mask:
            The single cell mask
        cue_direction:
            The orientation of the cue (e.g. flow)
        major_axes_length:
            The major axes length of the single cell
        num_partitions:
            The number of desired partitions

    Returns:
        The list of partitioned masks counter clock wise from the cue direction

    """
    # cv2 needs flipped y-axis
    sc_mask_f = np.flip(sc_mask, axis=0)

    # get the contour of the single cell mask
    contours = get_contour(sc_mask_f.astype(int))
    pg = Polygon(contours)
    pg_convex = pg.convex_hull

    pg_cent_a = int(pg.centroid.coords.xy[0][0])
    pg_cent_b = int(pg.centroid.coords.xy[1][0])

    # divisor line
    a = [pg_cent_a, pg_cent_b]
    b = [pg_cent_a + int(major_axes_length), pg_cent_b]  # lies horizontally
    div_line = LineString([a, b])

    # determine number of divisions
    divisors, div_angle = get_divisor_lines(a, cue_direction, div_line, num_partitions)

    div_coords = []
    for div in divisors:
        div_coords += [(int(np.rint(x)), int(np.rint(y))) for x, y in [*div.coords]]

    splits = LineString(div_coords)
    sectors = split(pg_convex, splits)

    polygons = list(sectors.geoms)

    # sort polygons counter clock wise
    polygons.sort(
        key=lambda x:
        compute_ref_x_abs_angle_deg(
            pg_cent_a, pg_cent_b,
            x.centroid.coords.xy[0][0], x.centroid.coords.xy[1][0]
        ) - (div_angle / 2) + cue_direction
    )

    masks = []
    for s in polygons:
        c = s.exterior.coords.xy
        x = np.asarray(c[0].tolist()).astype(np.uint)
        y = np.asarray(c[1].tolist()).astype(np.uint)

        # flip back to original y axis
        convex_mask = np.flip(mask_from_contours(sc_mask_f, x, y), axis=0)

        # keep only concave hull parts of the mask
        masks.append((np.logical_and(convex_mask, sc_mask)).astype(np.uint8))

    assert len(masks) == num_partitions, "Number of partitions(%s) does not match the number of masks (%s)." % (
        num_partitions, len(masks))

    return masks, polygons


def get_divisor_lines(origin, cue_direction, div_line, num_partitions):
    """Rotates a line from an origin point on the line based on a cue direction as often as specified

    Args:
        origin:
            origin point on the line
        cue_direction:
            The orientation of the cue (e.g. flow)
        div_line:
            shapely line object
        num_partitions:
            The number of desired partitions

    Returns:
        The list of partitioned masks counter clock wise from the cue direction

    """
    div_angle = int(360 / num_partitions)
    cumulative_angle = cue_direction + int(div_angle / 2)
    divisors = []
    while cumulative_angle < 360:
        divisors.append(rotate(div_line, cumulative_angle, origin=origin))
        cumulative_angle += div_angle
    return divisors, div_angle
